# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['esovalue']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.1.1', 'gmpy', 'mpmath']

setup_kwargs = {
    'name': 'esovalue',
    'version': '0.1.12',
    'description': 'Valuation of employee stock options',
    'long_description': '# ESO-value\nReceiving stock options from your company? Wondering what they are worth? ESO-value calculates the value of\nEmployee Stock Options based on the Hull-White model[1][2].\nThis library is used in the [ESO valuation app](https://eso.stephanethibaud.xyz/).\n\n## Installation\nRun `pip install esovalue`\n\n## Usage\n```python\nfrom esovalue.eso import value_eso\n\nvalue_eso(strike_price=50, stock_price=50, volatility=0.3, risk_free_rate=0.075,\n          dividend_rate=0.025, exit_rate=0.03, vesting_years=3, expiration_years=10, \n          iterations=1000, m=3)\n```\nDescription of the parameters:\n```\nstrike_price\t\t- Strike price\nstock_price\t\t- Current price of the underlying stock\niterations\t\t- More iterations is more precise but requires more memory/CPU\nrisk_free_rate\t\t- Risk-free interest rate\ndividend_rate\t\t- Dividend rate\nexit_rate\t\t- Employee exit rate (over a year)\nvesting_years\t\t- Vesting period (in years)\nexpiration_years\t- Years until expiration\nvolatility\t\t- Volatility (standard deviation on returns)\nm\t\t\t- Strike price multiplier for early exercise \n                          (exercise when the strike_price*m >= stock_price)\n```\n\n## References\n[1]: Hull, J, and White, A:  How to Value Employee Stock Options Financial Analysts Journal, Vol. 60, No. 1,\n    January/February 2004, 114-119.\\\n[2]: Hull, J. (2018). Options, Futures, and Other Derivatives (Global Edition).\n',
    'author': 'Stéphane Thibaud',
    'author_email': 'snthibaud@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/snthibaud/esovalue',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



from math import sqrt, e
from typing import Optional
from mpmath import mpf, mp
from esovalue.trinomial_tree import get_trinomial_tree, set_stock_prices, calculate_eso_prices
mp.dps = 100


def make_option_tree(strike_price: mpf, stock_price: Optional[mpf], volatility: mpf, risk_free_rate: mpf,
                     dividend_rate: mpf, exit_rate: mpf, vesting_years: mpf, expiration_years: mpf,
                     iterations: int, m: Optional[mpf]) -> mpf:
    strike_price, stock_price, volatility, risk_free_rate, dividend_rate, exit_rate, vesting_years, expiration_years, \
        m = [mpf(v) if v else v
             for v in [strike_price, stock_price, volatility, risk_free_rate, dividend_rate,
                       exit_rate, vesting_years, expiration_years, m]]
    dt = expiration_years / iterations
    root = get_trinomial_tree(iterations + 1)
    set_stock_prices(strike_price if stock_price is None else stock_price, e ** (volatility * sqrt(3 * dt)), root)
    calculate_eso_prices(root, strike_price, dt, volatility, risk_free_rate, dividend_rate, exit_rate, vesting_years,
                         m)
    return root


def value_eso(strike_price: mpf, stock_price: Optional[mpf], volatility: mpf, risk_free_rate: mpf,
              dividend_rate: mpf, exit_rate: mpf, vesting_years: mpf, expiration_years: mpf,
              iterations: int, m: Optional[mpf]) -> mpf:
    """
    Calculate the final value of an employee stock option (assumed 'at the money' if no stock price is given)
    :param strike_price: Strike price
    :param stock_price: Current price of the underlying stock
    :param iterations: More iterations is more precise but requires more memory/CPU
    :param risk_free_rate: Risk-free interest rate
    :param dividend_rate: Dividend rate
    :param exit_rate: Employee exit rate (over a year)
    :param vesting_years: Vesting period (in years)
    :param expiration_years: Years until expiration
    :param volatility: Volatility (standard deviation on returns)
    :param m: Strike price multiplier for early exercise (exercise when the strike_price*m >= stock_price)
    :return: Value in same currency as given strike price
    """
    return make_option_tree(strike_price, stock_price, volatility, risk_free_rate, dividend_rate, exit_rate,
                            vesting_years, expiration_years, iterations, m).option_value



from dataclasses import dataclass
from typing import Optional
from mpmath import mpf, ln, mp

ONE = mpf("1")
ZERO = mpf("0")


@dataclass
class TrinomialNode:
    stock_value: Optional[mpf] = None
    option_value: Optional[mpf] = None
    up: Optional['TrinomialNode'] = None
    middle: Optional['TrinomialNode'] = None
    down: Optional['TrinomialNode'] = None


def get_trinomial_tree(depth: int) -> TrinomialNode:
    assert depth > 0
    lattice = [[TrinomialNode()]]
    for level in range(1, depth):
        lattice.append([TrinomialNode() for _ in range(-level, level+1)])
    for i in range(0, len(lattice)-1):
        width = len(lattice[i])
        for j in range(0, width):
            lattice[i][j].down = lattice[i+1][j]
            lattice[i][j].middle = lattice[i+1][j+1]
            lattice[i][j].up = lattice[i+1][j+2]
    return lattice[0][0]


def set_stock_prices(s0: mpf, u: mpf, root: TrinomialNode):
    q = [(s0, root, True, True)]
    while q:
        s, n, h, l = q.pop()
        if n is not None:
            n.stock_value = s
            if h:
                q.append((s*u, n.up, True, False))
            if l:
                q.append((s/u, n.down, False, True))
            q.append((s, n.middle, False, False))


def present_value(fv: mpf, dt: mpf, r: mpf):
    """
    Calculate present value
    :param fv: Future value
    :param dt: Time difference (in years)
    :param r: Risk-free interest rate
    :return: Present value
    """
    return mp.e**(-r*dt)*fv


def calculate_eso_prices(root: TrinomialNode, k: mpf, dt: mpf, s: mpf, r: mpf, q: mpf, er: mpf, v: mpf,
                         m: Optional[mpf]):
    """
    Calculate the price of an employee stock option over time
    :param root: Root node
    :param k: Strike price
    :param dt: Length of one time step (in years)
    :param s: Volatility (standard deviation)
    :param r: Risk-free interest rate
    :param q: Dividend rate
    :param er: Employee exit rate (over a year)
    :param v: Vesting period (in years)
    :param m: Multiplier for early exercise
    """
    level = [root]
    levels = [level]
    while level:
        children = []
        for i, n in enumerate(level):
            middle_child = n.middle
            if middle_child:
                if i == 0:
                    children.append(n.down)
                children.append(middle_child)
                if i + 1 == len(level):
                    children.append(n.up)
            else:
                break
        level = children
        if level:
            levels.append(level)
    leaves = levels[-1]
    total_steps = len(levels)
    for n in leaves:
        n.option_value = max(ZERO, n.stock_value - k) if (total_steps-1)*dt >= v else ZERO
    for i in range(len(levels)-2, -1, -1):
        for node in levels[i]:
            vested = i * dt >= v
            if vested and m and node.stock_value >= k*m:
                node.option_value = node.stock_value - k
            else:
                a = mp.sqrt(dt/(12*s**2))*(r-q-s**2/2) if s > ZERO else ZERO
                b = ONE/6
                pd = -a + b
                pm = 4*b
                pu = a + b
                er_dt = ln(ONE+er)*dt
                option_value = (ONE-er_dt)*present_value(
                    pd*node.down.option_value + pm*node.middle.option_value + pu*node.up.option_value, dt, r)
                if vested:
                    option_value += er_dt*max(node.stock_value-k, ZERO)
                node.option_value = option_value
